{ ****************************************************************
  Sourcefile :  	bankContoll.pas
  Typ :  		controller-Unit
  Autor :  		max kleiner, LoC 406
  Beschreibung :  	handles observer and user event				fgung
  Klassen :	  	siehe ModelMaker
  Besonderes : 	        cooperates with bankLogic
  Revisionen :  	07.12.99 observer and dispatcher
			02.02.01 inbuilt password externals
                        05.06.04 findkey and refresh solved
                        15.06.04 clipboard and autoSave bank
                        17.06.04 refactoring zinsview to bview
 **************************************************************** }

unit bankControll;

interface

 uses bankLogic, classes, stdCtrls;
// ist ein versuch zur besseren modularisierung, parametrisierung
// bildet eine event-handler bridge zwischen logic und view
// entspricht diversen Pattern, somit view und logic besser austauschbar

type
  TObservable = class;
  // Observer abstract class interface
  TObserver = class
  protected
    procedure reset; virtual; abstract;
  public
    procedure Update(dispCenter: TObservable); virtual; abstract;
  end;

  // Subject abstract class interface
  TObservable = class
  private
    procedure allReset; virtual; abstract;
    procedure Notify; virtual; abstract;
  public
    procedure Add(Observer: TObserver); virtual; abstract;
    procedure Remove(Observer: TObserver); virtual; abstract;
  end;

  {TObsWrapper = class (TObserver)
  private
     FSample: TSample;
  protected
     function GetSomeValue: Integer;
     procedure SetSomeValue(Value: Integer);
  public
     function SomeAction(const Data: string): Boolean;
     property Sample: TSample read FSample;
     property SomeValue: Integer read GetSomeValue write SetSomeValue;
  end; }

  TDispCenter = class(TObservable)
  private
    FObservers: TList;
    FAmount: double;
    procedure Notify; override;
    procedure allReset; override;
    procedure changeAmount(value: double);
  public
    constructor Create;
    destructor Destroy; override;
      //TObservable
    procedure Add(Observer: TObserver); override;
    procedure Remove(Observer: TObserver); override;
      //Daten als Modell
    function getAmount: double;
  end;


type
 TCtrClass = class
 private
    FCustNo: integer;
    StGespeichert: boolean;
    StKontenAktiviert: boolean;
    FNormCharges: TChargeContext;
    FVorzugCharges: TChargeContext;
    FJugendCharges: TChargeContext;
 protected
   function canChange: boolean;
   procedure createKonten;
   procedure setKundenNo(custNo: integer);
 public
 //fixe instanzen der konten
   kontoClass: TKonto;
   stockKonto: TStockKonto;
   kreditKonto: TkreditKonto;
 //observer
   dispCtr: TDispCenter;
   tempList: TStringList;
   constructor CtrCreate;
   destructor Destroy; override;
   procedure ctrAction(konto: TKonto;
                         action: byte; amount: string);
   procedure toFile(dataList: TStringList);
   procedure SaveToBase;
   procedure ctrMutationen(tableID: byte);
   procedure createFactory;
   procedure copyToClipboard(aMemo: TMemo);
   function ctrStream(reportList: TStringList): TStringList;
   function FromFile(reportList: TStringList): TStringList;
   function getCustList: TStringList;
   function getVermoegen: double;
   property KundenNo: integer read FCustNo write setKundenNo;
 end;

 //instanz auf controller
 var ctr: TCtrClass;

 //password DLL, password in ulooktrans als konstante einstellbar
function GetPassword(const PassWord: string):boolean; far;
  external 'Password.dll';
//function SelectDir(dest: PChar): boolean; far;
//  external 'Password.dll';


implementation

uses SysUtils, bankView2, bankData, ulooktrans,
                         dialogs, controls;


constructor TDispCenter.Create;
begin
  inherited Create;
  FObservers:= TList.Create;
end;

destructor TDispCenter.Destroy;
begin
  FObservers.Free;
  inherited Destroy;
end;

procedure TDispCenter.Add(Observer: TObserver);
begin
  FObservers.Add(Observer);
end;

procedure TDispCenter.Remove(Observer: TObserver);
begin
  FObservers.Remove(Observer);
end;

procedure TDispCenter.Notify;
var  i: integer;
begin
  for i:= 0 to pred(FObservers.Count) do
    TObserver(FObservers.Items[i]).Update(Self);;
end;

procedure TDispCenter.allReset;
var  i: integer;
begin
  for i:= 0 to pred(FObservers.Count) do
    TObserver(FObservers.Items[i]).reset;
end;

function TDispCenter.getAmount: double;
begin
  result:= FAmount;
end;

procedure TDispCenter.changeAmount(value: double);
begin
  if value <> 0  then begin
    FAmount:= value;
  //state change notification
    Notify;
  end;
end;


constructor TCtrClass.CtrCreate;
begin
  inherited create;
  tempList:= TStringList.create;
  BusObj:= TBusObj.createBusObj;
  dispCtr:= TDispCenter.create;
  FNormCharges:= TJahrCharges.Create(TNormCharge.Create);
  FVorzugCharges:= TJahrCharges.Create(TVorzugCharge.Create);
  FJugendCharges:= TJahrCharges.Create(TJugendCharge.Create);
end;

destructor TCtrClass.Destroy;
begin
  KontoClass.free;
  FNormCharges.Free;
  FVorzugCharges.Free;
  FJugendCharges.Free;
  tempList.free;
  dispCtr.free;
  BusObj.Free;
  inherited Destroy;
end;

function TCtrClass.canChange: boolean;
begin
  result:= messageDlg(frmTrans.strLit[20],mtConfirmation,
                           [mbOK, mbCancel],0) = mrOK;
end;

function TCtrClass.getCustList: TStringList;
begin
  result:= busObj.getAllCustNo;
end;

function TCtrClass.getVermoegen: double;
begin
  result:= busObj.getAllVermoegen(FCustNo);
end;

procedure TCtrClass.setKundenNo(custNo: integer);
begin
  if custNo > CLIENT_BASE then
                   FCustNo:= custNo;
end;


// konto, business process(action) und betrag via parameters
{ /1 kreditlimite /2 buchen /3 umsatz /4 persistenz in file
  /5 gebuehren cases /6 speichern konto in bank /7 ist reserve
  /8 transaktion history}
//**************************** action event handler ***********
procedure TCtrClass.
     ctrAction(konto: TKonto; action: byte; amount: string);
begin
  case action of
  1: begin
      TKonto(konto).KreditLimit:= StrToFloat(amount);
     end;
  //buchen
  2: begin
       TKonto(konto).Kontostand:= StrToFloat(amount);
       dispCtr.changeAmount(strToFloat(amount));
       if assigned(transmonitor)
          then transmonitor.addTransactions(amount);
     end;
  3: begin
  //calc aggregation
      TKonto(konto).Transakt_Obj.Umsatz;
      tempList.clear;
      tempList.addstrings(TKonto(konto).getData);
      tempList.addstrings(TKonto(konto).Transakt_Obj.getData);
     end;
  4: begin
        busObj.writeToFile((TKonto(konto).getData));
      end;
  5: begin
       if konto is TStockkonto then
        TStockkonto(konto).Depotgebuehr:= StrToFloat(amount);
     end;
  51: begin
       TKreditkonto(konto).Jahresgebuehr:=
           FNormCharges.ComputeCharges(StrToFloat(amount));
      end;
  52: begin
       TKreditkonto(konto).Jahresgebuehr:=
          FVorzugCharges.ComputeCharges(StrToFloat(amount));
      end;
  53: begin
       TKreditkonto(konto).Jahresgebuehr:=
          FJugendCharges.ComputeCharges(StrToFloat(amount));
      end;
  // save in bank
  6: begin
        TKonto(konto).saveData;
     end;
  //7: begin
  //reserve
  //   end;
  8: begin
  // transaktionen kontoreport
       tempList.clear;
       tempList.addStrings(TKonto(konto).
                              Transakt_Obj.TrsaktHistory);
     end;
   end;
 end;

//********************** mutationen auf bankView2 *************
procedure TCtrClass.ctrMutationen(tableID: byte);
  var onlyInst: TFrmDataView2;
  //to get the classMethod
  //oldHeight: integer;
  begin
  case tableID of
  //singleton pattern
   1:begin
     onlyInst:= TFrmDataView2.onlyInstance;
       with onlyInst do begin
         show;
         height:= oldHeight;
         dataModule1.tblCust.close;
         dataModule1.tblCust.Open;
         dataModule1.tblCust.findkey([kundenno]);
         DBGrid1.dataSource:= DataModule1.dSrcCust;
         dbGrid1.ReadOnly:= false;
         dbGrid2.dataSource:= dataModule1.dSrcAccount;
         dbGrid2.visible:= true;
         dbGrid2.ReadOnly:= false;
         DBNavigator1.dataSource:= DataModule1.dSrcCust;
       end;
     end;
   //kontotyp daten
   2:begin
      onlyInst:= TFrmDataView2.onlyInstance;
      with onlyInst do begin
        show;
        height:=newHeight;
        DBGrid1.dataSource:= DataModule1.dSrcAccType;
        dbGrid1.ReadOnly:= false;
        DBGrid2.visible:= false;
        DBNavigator1.dataSource:= DataModule1.dSrcAccType;
      end;
     end;
   3:begin
      onlyInst:=TFrmDataView2.onlyInstance;
      with dataModule1.tblAccount do begin
      busObj.getSQLTransHist(fieldbyname('CUST_NO').asInteger,
                            fieldbyName('ACC_TYPE').asInteger);
      end;
      //transaktionsdaten
      with onlyInst do begin
        show;
        height:= newHeight;
        DBGrid1.dataSource:= dataModule1.dSrcTransHist;
        DBGrid2.visible:= false;
        dbGrid1.ReadOnly:= true;
        DBNavigator1.dataSource:= dataModule1.dSrcTransHist;
      end;
    end;
  end;
end;

//alle konten gespeichert
procedure TCtrClass.SaveToBase;
begin
  kontoClass.saveData;
  stockKonto.saveData;
  kreditKonto.saveData;
  StGespeichert:= true;
end;

procedure TCtrClass.ToFile(dataList: TStringList);
begin
//write all memo to file
  busObj.writeToFile(dataList);
end;

function TCtrClass.FromFile(reportList:TStringList): TStringList;
begin
  result:= busObj.readFromFile(reportList);
end;

function TCtrClass.ctrStream(reportList:TStringList): TStringList;
begin
  result:= busObj.getStreamTemplate(reportList);
end;

// instanziert die objekte via kundennummer
{aus vereinfachung wird als nummer direkt die kundennummer
bergeben, in real wrde ein bank-objekt die einzelnen
kontoinstanzen in einem constructor-loop in die welt setzen}

procedure TctrClass.createKonten;
 var std_account: byte;
begin
//ist der account_typ;
  std_account:= SPAR_KTO;
  kontoClass:= TKonto.createKonto(FCustNo, std_account);
  stockKonto:= TStockKonto.createKonto(FCustNo);
  kreditKonto:= TKreditKonto.createKonto(FCustNo);
  assert(frmTrans.checkOBJ(kontoClass), frmTrans.strLit[41]);   //trans
  StKontenaktiviert:= True;
  StGespeichert:= false;
end;

procedure TCtrClass.createFactory;
 begin
   if assigned(kontoClass) then begin
    //if (canChange and StGespeichert) then begin
     if canChange then begin
       savetoBase;
       kontoClass.free;
       dispCtr.allReset;
       createKonten;
     end;
   // erstmalig
   end else createKonten;
 end;

procedure TCtrClass.copyToClipboard(amemo: TMemo);
begin
  if amemo.Lines.Count > 1 then begin
    //for i:= 1 to amemo.Lines.Count - 1 do
    amemo.SelectAll;
    amemo.CopyToClipboard;
  end;
end;

end.
