{ ****************************************************************
  Sourcefile :  	bankLogic.pas
  Typ :  		Business-Unit
  Autor :  		max kleiner, LoC=462
  Beschreibung :  	Zeigt die Bankprozesse/ Gebhrenstrategie				fgung
  Klassen :	  	siehe ModelMaker
  Besonderes : 	        kooperiert mit bankControll
  Revisionen :  	07.11.99 mit Basisklassen der Konten
			02.02.00 Gebhrenstrategie
                        05.06.04 FCustName typed of string
 **************************************************************** }

unit bankLogic;

interface

uses classes, SysUtils, bankData, db ;  //TString in classes

type
 //forward declaration, because of aggregation
 TKonto = class;
 TTransaction = class(TBusObj)
  private
    FPosten: integer;
    FMonat: byte;
    FTurnover: double;
    FTurnoverAvg: double;
    FKonto_Obj: TKonto;
    dataList: TStringList;
  protected
    procedure SetTransactionHist(betrag: double);
  public
    constructor createTrans(konto: TKonto);
    destructor Destroy; override;
    function Umsatz: double;
    function TrsaktHistory: TStringList;
    function getData: TStringList;
  end;


  TKonto = class(TBusObj)
  private
    dataList: TStringList;
    FID: integer;
    FCustName: string[240];
    FStammCode: longInt;
    FKreditLimit: double;
    FBasisGebuehr: double;
    FKontostand: double;
    FDiffKonto: byte;
    FTransakt_Obj: TTransaction; //aggregation
  protected
    function getKreditLimit: double;
    procedure setKreditLimit(newLimit: double);
    procedure Buchen(betrag: double); virtual;
    function collectData: TStringList; virtual;
  public
    constructor createKonto(cust_No: longInt; std_acc: byte);
    destructor Destroy; override;
    function getData: TStringList;
    procedure saveData; virtual;
    property KreditLimit: double
      read getKreditLimit write setKreditLimit;
    property Kontostand: double
      read FKontostand write Buchen;
    property Transakt_Obj: TTransaction
      read FTransakt_Obj;
  end;


 TStockKonto = class(TKonto)
  private
    FDepotGebuehr: double;
  protected
    procedure Buchen(betrag: double); override;
    procedure setDepotGebuehr(neueBasis: double);
    function collectData: TStringList; override;
  public
    constructor createKonto(cust_No: longint);
    destructor Destroy; override;
    procedure saveData; override;
    property DepotGebuehr: double
      read FDepotGebuehr write setDepotGebuehr;
 end;

 TkreditKonto = class(TKonto)
  private
    FJahresGebuehr: double;
  protected
    function collectData: TStringList; override;
    function Bonitaet(betrag: double): boolean;
    procedure Buchen(betrag: double); override;
    procedure setJahresGebuehr(neueGebuehr: double);
  public
    constructor createKonto(cust_No: longint);
    destructor Destroy; override;
    procedure saveData; override;
    property JahresGebuehr: double
      read FJahresGebuehr write setJahresGebuehr;
 end;

  type
    //strategie interface fuer gebuehren des kreditkonto
  TKreditCharge = class
  public
    //gibt die gebuehren belastung zurck
    function getCharge(const Balance: double): double; virtual; abstract;
  end;

    //konkrete strategie
  TNormCharge = class(TKreditCharge)
  public
    function getCharge(const Balance: double): double; override;
  end;

  TVorzugCharge = class(TKreditCharge)
  public
     function getCharge(const Balance: double): double; override;
  end;

  TJugendCharge = class(TKreditCharge)
  public
    function getCharge(const Balance: double): double; override;
  end;

 //Context Interface
  TChargeContext = class
  public
    function ComputeCharges(const Balance: double): double;
                                               virtual; abstract;
  end;

  //Concrete Context
  TJahrCharges = class(TChargeContext)
  private
    FKreditCharge: TKreditCharge;
  public
    function ComputeCharges(const Balance: double): double; override;
    constructor Create(aKreditCharge: TKreditCharge); virtual;
    destructor Destroy; override;
  end;


implementation

uses dialogs, ulooktrans, bankcontroll;

 constructor TTransaction.createTrans(konto: TKonto);
 begin
  //aggregation
    FKonto_Obj:= konto;
  //tagesumsatz
    FPosten:= 0;
    FMonat:= 1;
    dataList:= TStringList.create;
 end;

constructor TKonto.createKonto(cust_No: longInt;
                                          std_Acc: byte);
  begin
   //if(dataModule1.q_data.active) then begin
   // check to close the account
   //end;
   //legt daten in dataset
    getSQLKonto(cust_No, std_Acc);
  try
    with dataModule1.dsOpen.dataSet do begin
      FStammCode:= fieldValues['ACC_NO'];
      FDiffKonto:=fieldValues['ACC_TYPE'];
      FID:= fieldValues['CUST_NO'];
      FCustName:= fieldValues['CUSTOMER'];
      FKreditLimit:= fieldValues['LIMIT'];
      FKontostand:= fieldValues['AMOUNT'];
      FBasisGebuehr:= fieldValues['FEE'];
    end;
  except on EVariantError do begin
      raise Exception.create(frmTrans.strLit[13]);
    end;
  end;
  //for the collector
   dataList:= TStringList.create;
    FTransakt_Obj:= TTransaction.createTrans(self);
  end;

  constructor TStockKonto.createKonto(cust_No: longInt);
  begin
    inherited createKonto(cust_No, STOCK_KTO);
    with dataModule1.dsOpen.dataSet do begin
      FDepotGebuehr:= fieldValues['FEE'];
    end;
  end;

  constructor TkreditKonto.createKonto(cust_No: longInt);
  begin
    inherited createKonto(cust_No, KRED_KTO);
    with dataModule1.dsOpen.dataSet do begin
      FJahresGebuehr:= fieldValues['FEE'];
    end;
  end;

  procedure TKonto.setKreditLimit(newLimit: double);
  begin
    if FKreditLimit <> newLimit then begin
      if getPassword(frmTrans.password) then begin
         FKreditLimit:= newLimit
      end else MessageDlg(frmTrans.strLit[2]+'  '+
             frmTrans.strLit[3], mtWarning,[mbok],0);
    end;
  end;

  function TKonto.getKreditLimit;
  begin
    result:= FKreditLimit;
  end;

  procedure TKonto.Buchen(betrag: double);
  begin
    FKontostand:= FKontostand + Betrag;
    FTransakt_Obj.setTransactionHist(betrag);
  end;

 procedure TStockKonto.Buchen(betrag: double);
  begin
    inherited Buchen(betrag);
    // abziehen vom sparkonto
    ctr.KontoClass.FKontoStand:=
               ctr.kontoClass.FKontoStand - Betrag;
  end;

 procedure TkreditKonto.Buchen(betrag: double);
  begin
  if betrag > 0 then
    inherited Buchen(betrag)
  else if Bonitaet(betrag) then
    inherited Buchen(betrag)
    else  MessageDlg(frmTrans.strLit[10], mtWarning,[mbok],0);
  end;

 procedure TKonto.saveData;
 begin
   setSQLKonto(FStammCode);
   with dataModule1 do begin
     if dsOpen.dataSet.active then
       dataBeforePost(dsOpen.dataSet);
       dsOpen.dataSet.fieldvalues['AMOUNT']:= Kontostand;
       dsOpen.dataSet.fieldvalues['LIMIT']:= KreditLimit;
       dsOpen.dataSet.fieldvalues['FEE']:= FBasisGebuehr;
       dataAfterPost(dsOpen.dataSet);
     end;
  end;

 procedure TStockKonto.saveData;
 begin
   inherited saveData;
   with dataModule1 do begin
    dataBeforePost(dsOpen.dataSet);
    dsOpen.DataSet.fieldvalues['FEE']:= FDepotGebuehr;
    dataAfterPost(dsOpen.dataSet);
    end;
 end;

 procedure TKreditKonto.saveData;
  begin
   inherited saveData;
   with dataModule1 do begin
     dataBeforePost(dsOpen.dataSet);
     dsOpen.DataSet.fieldvalues['FEE']:= FJahresGebuehr;
     dataAfterPost(dsOpen.dataSet);
   end;
  end;

 function TKonto.collectData: TStringList;
  begin
    with dataList do begin
      clear;
      add(''); add('KontoNr: ' + intToStr(FStammCode)
        +'-'+(intToStr(FDiffKonto)));
      add('Kontoinhaber: '+ FCustName);
      add('Kunden Nr.: ' +intToStr(FID ));
      add('Kredit Limite: ' + FloatToStr(FKreditLimit));
      add('Kontostand: ' + FloatToStr(Kontostand));
    end;
   result:= dataList;
  end;

 function TStockKonto.collectData: TStringList;
  begin
    with dataList do begin
      inherited collectData;
      add('Depot Gebuehr: '+ FloattoStr(FDepotGebuehr));
    end;
   result:= dataList;
  end;

  function TkreditKonto.collectData: TStringList;
  begin
    with dataList do begin
      inherited collectData;
      add('Jahres Gebuehr: '+floatToStr(FJahresGebuehr));
    end;
   result:= dataList;
  end;

  //virtual container
  function TKonto.getData: TStringList;
  begin
  //collectData is polymorphic
    result:= collectData;
  end;


  function TkreditKonto.Bonitaet(betrag: double): boolean;
  begin
    if (KontoStand - -Betrag) >= -FKreditLimit then begin
     //x:= KontoStand - -Betrag; debug code...
     result:= true end else result:=false;
  end;

  procedure TkreditKonto.setJahresGebuehr(neueGebuehr: double);
  begin
    FJahresGebuehr:= 0;
    FJahresGebuehr:= neueGebuehr;
  end;

  procedure TStockkonto.setDepotGebuehr(neueBasis: double);
  begin
    FDepotGebuehr:= 0;
    FDepotGebuehr:= neueBasis +
    (ctr.kontoClass.FBasisGebuehr * 4);
  end;

// TTransactionClass
procedure TTransaction.SetTransactionHist(betrag: double);
begin
  if frmTrans.boolLogSave then begin
   if frmTrans.chkBox_Stat.Checked then
     setSP_TransData(betrag, FKonto_Obj.FStammCode) else
     setSQLTransData(betrag, FKonto_Obj.FStammCode);
  end;
 inc(FPosten);
end;

function TTransaction.Umsatz: double;
begin
  //aggregation & friend (private!)
  FTurnover:= getSQLUmsatz(FKonto_Obj.FID, FKonto_Obj.FStammCode);
  try
    FTurnoverAvg:= (FKonto_Obj.FKontostand / FPosten);
  except
    on EZeroDivide do begin
      //FTurnoverAvg := 1;
      showmessage(frmTrans.strLit[5])
    end;
  end;
 result:= FTurnover;
end;

function TTransaction.getData: TStringList;
  begin
    with dataList do begin
      clear;
      add('');
      add('Monate: '+floatToStr(FMonat));
      add('Posten Transaktion: '+ FloattoStr(FPosten));
      add('Tagesumsatz: '+ floatToStr(round(FTurnover)));
      add('Umsatz Durchschnitt: '+ floatToStr(round(FTurnoverAvg)));
    end;
   result:= dataList;
  end;

function TTransaction.TrsaktHistory: TStringList;
begin
  getSQLTransHist(FKonto_Obj.FID, FKonto_Obj.FDiffKonto);
  with dataList do begin
    clear;
    with dataModule1.dSrcTransHist do begin
      while not dataSet.EOF do begin
        add(' ');
        add(dataSet.fieldValues['ACC_NO']);
        add(dataSet.fieldValues['TRANSAMOUNT']);
        add(dataSet.fieldByName('OPENDATE').asString);
        dataSet.next;
      end;
    end;
  end;
    result:= dataList
end;

//**************garbage collection*************************
destructor TKonto.Destroy;
  begin
    ctr.stockKonto.free;
    ctr.kreditKonto.free;
    dataList.Free;
    Transakt_Obj.Free;
    ctr.kontoClass:= NIL;
    FTransakt_Obj:= NIL;
    inherited Destroy;
  end;

destructor TStockKonto.Destroy;
  begin
    dataList.Free;
    Transakt_Obj.Free;
    ctr.stockKonto:= NIL;
  end;

destructor TKreditKonto.Destroy;
  begin
    dataList.Free;
    Transakt_Obj.Free;
    ctr.kreditKonto:= NIL;
  end;

destructor TTransaction.Destroy;
  begin
    dataList.free;
    //inherited destroy;
  end;

{********Strategy Pattern*************************************}
// TNormalCharge
function TNormCharge.getCharge(const Balance: double): double;
begin
  result:= Balance * (REG_RATE * 12);
end;

// TVorzugCharge
function TVorzugCharge.getCharge(const Balance: double): double;
begin
  //creditcard holders buying patterns rewards take into account
  result:= Balance * (PREFERED_RATE * 12);
end;

// TJugendCharge
function TJugendCharge.getCharge(const Balance: double): double;
begin
  result:= Balance * (TRIAL_RATE * 12);
end;

// Concrete Context
constructor TJahrCharges.Create(aKreditCharge: TKreditCharge);
begin
  inherited Create;
  if not assigned(aKreditCharge) then
    raise Exception.Create(frmTrans.strLit[4]);
    FKreditCharge:= aKreditCharge;
end;

destructor TJahrCharges.Destroy;
begin
  FKreditCharge.Free;
  inherited Destroy;
end;

 // context interface seen by client objects
function TJahrCharges.ComputeCharges(const Balance: double): double;
begin
  result:= FKreditCharge.getCharge(Balance);
end;

end.
